/* 
   ADV7175A - Analog Devices ADV7175A video encoder driver version 0.0.3

   Copyright (C) 2000 Henrik Johannson <henrikjo@post.utfors.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <linux/types.h>
#include <linux/wrapper.h>

#include <linux/videodev.h>
#include <linux/version.h>
#include <asm/uaccess.h>

#include <linux/i2c-algo-bit.h>
#include <linux/video_encoder.h>

#include "adv7175.h"

#define i2c_is_isa_client(clientptr) \
        ((clientptr)->adapter->algo->id == I2C_ALGO_ISA)
#define i2c_is_isa_adapter(adapptr) \
        ((adapptr)->algo->id == I2C_ALGO_ISA)


#define ADV7175_REG_MR0 0
#define ADV7175_REG_MR1 1
#define ADV7175_REG_TTXRQ_CTRL 0x24

static int adv7175_attach_adapter(struct i2c_adapter *adapter);
int adv7175_detach_client(struct i2c_client *client);
int adv7175_command(struct i2c_client *client, unsigned int cmd, void *arg);
void adv7175_inc_use (struct i2c_client *client);
void adv7175_dec_use (struct i2c_client *client);

#define CHIP_ADV7175A 1
#define CHIP_ADV7170  2

#define MODE_PAL 1
#define MODE_NTSC 1

struct adv7175_data_s {
    int chiptype;
    int mode;
    int bars;
};

/* This is the driver that will be inserted */
static struct i2c_driver adv7175_driver = {
  /* name */            "ADV7175 video encoder driver",
  /* id */              I2C_DRIVERID_ADV7175,
  /* flags */           I2C_DF_NOTIFY,
  /* attach_adapter */  &adv7175_attach_adapter,
  /* detach_client */   &adv7175_detach_client,
  /* command */         &adv7175_command,
  /* inc_use */         &adv7175_inc_use,
  /* dec_use */         &adv7175_dec_use
};

#define   I2C_ADV7175        0x88

int adv7175_id = 0;

static unsigned char PAL_config[16] = {
    0x01,       // Mode Register 0
    0x6,        // Mode Register 1
    0xcb,       // Subcarrier Frequency Register 0
    0x8a,       // Subcarrier Frequency Register 1
    0x09,       // Subcarrier Frequency Register 2
    0x2a,       // Subcarrier Frequency Register 3
    0x00,       // Subcarrier Phase Register 
    0x4d,       // Timing Register 0
    0x00,       // Closed Captioning Ext Register 0
    0x00,       // Closed Captioning Ext Register 1
    0x00,       // Closed Captioning Register 0 
    0x00,       // Closed Captioning Register 1 
    0x73,       // Timing Register 1
    0x08,       // Mode Register 2
    0x00,       // Pedestal Control Register 0
    0x00,       // Pedestal Control Register 1
};

static
int adv7175_update(struct i2c_client *client) {
    struct adv7175_data_s *data = client->data ;
    int n,i;

    if(data->mode == MODE_PAL) {
	for(i=0; i<16; i++) 
	    n=i2c_smbus_write_byte_data(client,i,PAL_config[i]);
    }

    if(data->bars) 
	i2c_smbus_write_byte_data(client,1,i2c_smbus_read_byte_data(client,1) | 0x80);

    i2c_smbus_write_byte_data(client,7, i2c_smbus_read_byte_data(client,7) & 0x7f);
    i2c_smbus_write_byte_data(client,7, i2c_smbus_read_byte_data(client,7) | 0x80);
    i2c_smbus_write_byte_data(client,7, i2c_smbus_read_byte_data(client,7) & 0x7f);
    return 0;
}

static
int adv7175_setup(struct i2c_client *client) {
    struct adv7175_data_s *data = client->data ;

    data->mode=MODE_PAL;
    data->bars=0;
    adv7175_update(client);
    return 0;
}
static int adv7175_detect(struct i2c_adapter *adapter, int address)
{
    struct adv7175_data_s *data;
    struct i2c_client *new_client;
    int mr0,mr1;
    int err;

    if(i2c_is_isa_adapter(adapter)) {
	printk("adv7175a.o: called for an ISA bus adapter?!?\n");
	return 0;
    }

     if (! i2c_check_functionality(adapter,I2C_FUNC_SMBUS_READ_BYTE| 
                                        I2C_FUNC_SMBUS_WRITE_BYTE_DATA))
	 return 0;

     
     
     if (! (new_client = kmalloc(sizeof(struct i2c_client) +
				 sizeof(struct adv7175_data_s),
				 GFP_KERNEL))) 
	 return -ENOMEM;

     data = (struct adv7175_data_s *) (((struct i2c_client *) new_client) + 1);
     new_client->addr = address;
     new_client->data = data;
     new_client->adapter = adapter;
     new_client->driver = &adv7175_driver;
     new_client->flags = 0;

     i2c_smbus_write_byte_data(new_client,ADV7175_REG_MR1,0x55);
     mr1=i2c_smbus_read_byte_data(new_client,ADV7175_REG_MR1);
     
     if(mr1 == 0x55) {
	 strcpy(new_client->name, "ADV7175A chip");
	 new_client->id = adv7175_id++;

	 if ((err = i2c_attach_client(new_client))) {
	     kfree(new_client);
	     return err;
	 }

	 mr0=i2c_smbus_read_byte_data(new_client,ADV7175_REG_MR0);

	 if(mr0 & 0x20) {
	     data->chiptype = CHIP_ADV7175A;
	     printk("adv7175.o: ADV7175A chip detected\n");
	 } else {
	     data->chiptype = CHIP_ADV7170;
	     printk("adv7175.o: ADV7170 chip detected\n");
	 }
	 
	 adv7175_setup(new_client);

	 return 0;
     }

     
     kfree(new_client);
     return 0;              
}

static int adv7175_attach_adapter(struct i2c_adapter *adapter)
{
    adv7175_detect(adapter,0x6a);
    adv7175_detect(adapter,0x7a);
    adv7175_detect(adapter,0xa);
    return 0;
}


int adv7175_detach_client(struct i2c_client *client)
{
  int err;

  if ((err = i2c_detach_client(client))) {
    printk("adv7175.o: Client deregistration failed, client not detached.\n");
    return err;
  }

  kfree(client);

  return 0;
}

/* No commands defined yet */
int adv7175_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
    return 0;
}

/* Nothing here yet */
void adv7175_inc_use (struct i2c_client *client)
{
#ifdef MODULE
  MOD_INC_USE_COUNT;
#endif
}

/* Nothing here yet */
void adv7175_dec_use (struct i2c_client *client)
{
#ifdef MODULE
  MOD_DEC_USE_COUNT;
#endif
}

/* ----------------------------------------------------------------------- */


#ifdef MODULE
int init_module(void)
#else
int adv7175_init(void)
#endif
{
    return i2c_add_driver(&adv7175_driver);
}



#ifdef MODULE

void cleanup_module(void)
{
    i2c_del_driver(&adv7175_driver);
}

#endif
