/*
    em8300.c - EM8300 MPEG-2 decoder device driver

    Copyright (C) 2000 Henrik Johansson <henrikjo@post.utfors.se>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <linux/string.h>
#include <linux/time.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>

#include <linux/version.h>
#include <asm/uaccess.h>

#include <linux/i2c-algo-bit.h>

#include "em8300.h"

#include "em8300_reg.c"

MODULE_AUTHOR("Henrik Johansson <henrikjo@post.utfors.se>");
MODULE_DESCRIPTION("EM8300 MPEG-2 decoder");
MODULE_PARM(remap,"1-" __MODULE_STRING(EM8300_MAX) "i");

static int em8300_cards;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
static unsigned int remap[EM8300_MAX]={};
#endif
static struct em8300_s em8300[EM8300_MAX];

#define TIMEDIFF(a,b) a.tv_usec - b.tv_usec + \
	    1000000 * (a.tv_sec - b.tv_sec)

int em8300_waitfor(struct em8300_s *em, int reg, int val, int mask) {
    int tries;

    for(tries=0; tries < 100 ; tries++) {
	if(em->mem[reg] == val)
	    return 0;
	udelay(10000);
    }
    return -ETIME;
}

static
int mpegvideo_command(struct em8300_s *em, int cmd) {
    if(UCREGV(MV_Command) != 0xffff)
	return -1;

    UCREGV(MV_Command) = cmd;

    return em8300_waitfor(em,UCREG(MV_Command), 0xffff, 0xffff);
}

int mpegaudio_command(struct em8300_s *em, int cmd) {
    if(UCREGV(MV_Command) != 0xffff)
	return -1;

    UCREGV(MA_Command) = cmd;

    return em8300_waitfor(em,UCREG(MA_Status), cmd, 0xffff);
}

static
int dicom_update(struct em8300_s *em) {
    int ret;
    
    if( (ret=em8300_waitfor(em, UCREG(DICOM_UpdateFlag), 0, 1)) )
	return ret;

    UCREGV(DICOM_BCSLuma) = em->dicom.luma;
    UCREGV(DICOM_BCSChroma) = em->dicom.chroma;
    UCREGV(DICOM_FrameTop) = em->dicom.frametop;
    UCREGV(DICOM_FrameBottom) = em->dicom.framebottom;
    UCREGV(DICOM_FrameLeft) = em->dicom.frameleft;
    UCREGV(DICOM_FrameRight) = em->dicom.frameright;
    UCREGV(DICOM_VisibleTop) = em->dicom.visibletop;
    UCREGV(DICOM_VisibleBottom) = em->dicom.visiblebottom;
    UCREGV(DICOM_VisibleLeft) = em->dicom.visibleleft;
    UCREGV(DICOM_VisibleRight) = em->dicom.visibleright;
    UCREGV(DICOM_TvOut) = em->dicom.tvout;

    UCREGV(DICOM_UpdateFlag) = 1;
    
    return em8300_waitfor(em, UCREG(DICOM_UpdateFlag), 0, 1);
}

/*
  FIFO routines
*/

static
int em8300_init_fifo(struct em8300_s *em, struct fifo_s *f,
		     int start, int wrptr, int rdptr,
		     int pcisize, int slotsize, int fifotype)
{
    int i;
    unsigned phys;

    f->type = fifotype;
    
    f->writeptr = (unsigned *)&UCREGV(wrptr);
    f->readptr = (unsigned *)&UCREGV(rdptr);

    switch(f->type) {
    case FIFOTYPE_AUDIO:
	f->slotptrsize = 3;
	f->slots.a = (struct audio_fifoslot_s *) &UCREGV(start);
	f->nslots = UCREGV(pcisize) / 3;
	break;
    case FIFOTYPE_VIDEO:
	f->slotptrsize = 4;
	f->slots.v = (struct video_fifoslot_s *) &UCREGV(start);
	f->nslots = UCREGV(pcisize) / 4;
	break;
    }
    
    f->slotsize = slotsize;
    f->start = UCREG(start) - 0x1000;
    f->threshold = f->nslots / 2;
    f->waiting=0;

    if(f->fifobuffer)
	kfree(f->fifobuffer);
    
    f->fifobuffer = kmalloc(f->nslots * f->slotsize, GFP_KERNEL);
    if(f->fifobuffer == NULL)
	return -ENOMEM;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
    init_waitqueue(&f->wait);
#else
    init_waitqueue_head(&f->wait);
#endif	
    
    for(i=0; i < f->nslots; i++) {
	phys = virt_to_phys(f->fifobuffer + i*f->slotsize);
	switch(f->type) {
	case FIFOTYPE_AUDIO:	
	    f->slots.a[i].physaddress_hi = phys >> 16;
	    f->slots.a[i].physaddress_lo = phys & 0xffff;
	    f->slots.a[i].slotsize = f->slotsize;
	    break;
	case FIFOTYPE_VIDEO:
	    f->slots.v[i].flags = 0;
	    f->slots.v[i].physaddress_hi = phys >> 16;
	    f->slots.v[i].physaddress_lo = phys & 0xffff;
	    f->slots.v[i].slotsize = f->slotsize;
	}
    }

    f->valid = 1;
    
    return 0;
}

void em8300_free_fifo(struct fifo_s *f)
{
    if(f->valid) {
	if(f->fifobuffer) 
	    kfree(f->fifobuffer);
    }
    f->valid=0;
}

static
int em8300_write_fifo(struct fifo_s *fifo, int n, const char *userbuffer,
		      int swapbyteorder)
{
    int freeslots,writeindex,i,bytes_transferred=0,copysize;
    unsigned short *sptr;

    if(!fifo->valid)
	return -1;
    
    freeslots =
	((*fifo->readptr - *fifo->writeptr) / fifo->slotptrsize +
	 fifo->nslots - 1) % fifo->nslots;

    writeindex = (*fifo->writeptr - fifo->start) / fifo->slotptrsize;
    for(i=0; i < freeslots && n; i++) {
	copysize = n < fifo->slotsize ? n : fifo->slotsize;

	switch(fifo->type) {
	case FIFOTYPE_AUDIO:
	    fifo->slots.a[writeindex].slotsize = copysize;
	    break;
	case FIFOTYPE_VIDEO:
	    fifo->slots.v[writeindex].flags = 0x0;
	    fifo->slots.v[writeindex].slotsize = copysize;
	    break;
	}
	
	copy_from_user(fifo->fifobuffer + writeindex*fifo->slotsize,
		       userbuffer, copysize);

	if(swapbyteorder) {
	    sptr = (unsigned short *) (fifo->fifobuffer +
		                      writeindex*fifo->slotsize);
	    for(i=0; i < copysize / 2; i++) 
		sptr[i] = (sptr[i]>>8) | (sptr[i]<<8);
	}
	
	writeindex++;
	writeindex %= fifo->nslots;
	n -= copysize;
	userbuffer += copysize;
	bytes_transferred += copysize;
    }
    *fifo->writeptr = fifo->start + writeindex*fifo->slotptrsize;
    return bytes_transferred;
}


static
int em8300_check_fifo(struct fifo_s *fifo) {
    int freeslots;
    
    if(!fifo->valid)
	return -1;
    
    freeslots =
	((*fifo->readptr - *fifo->writeptr) / 4 + fifo->nslots - 1) %
	fifo->nslots;

    if((freeslots > fifo->threshold) && fifo->waiting) {
	wake_up_interruptible(&fifo->wait);
    }
    return 0;
}

void em8300_clockgen_write(struct em8300_s *em, int abyte) {
    int i;

    em->mem[EM8300_I2C_PIN] = 0x808;
    for(i=0; i < 8; i++) {
	em->mem[EM8300_I2C_PIN] = 0x2000;
	em->mem[EM8300_I2C_PIN] = 0x800 | ((abyte & 1) ? 8:0);
	em->mem[EM8300_I2C_PIN] = 0x2020;
	abyte >>= 1;
    }

    em->mem[EM8300_I2C_PIN] = 0x200;
    em->mem[EM8300_I2C_PIN] = 0x202;
}	

static
void em8300_irq(int irq, void *dev_id, struct pt_regs * regs)
{
    struct em8300_s *em = (struct em8300_s *)dev_id;
    int irqstatus;
    struct timeval tv;

    irqstatus = UCREGV(Q_IrqStatus);

    if( irqstatus & 0x8000) {
	UCREGV(Q_IrqMask) = 0x0;
	em->mem[0x1ffa] = 2;

	UCREGV(Q_IrqStatus) = 0x8000;

	if(irqstatus & IRQSTATUS_VIDEO_FIFO) {
	    em8300_check_fifo(&em->mvfifo);
	    do_gettimeofday(&tv);
	    em->irqtimediff = TIMEDIFF(tv,em->tv);
	    em->tv = tv;
	    em->irqcount++;
	}
	
	if(irqstatus & IRQSTATUS_AUDIO_FIFO) 
	    em8300_check_fifo(&em->mafifo);
	
	UCREGV(Q_IrqMask) = em->irqmask;
	UCREGV(Q_IrqStatus) = 0x0000;
    }
}

static
void release_em8300(int max)
{
    struct em8300_s *em;
    int i;

    for (i=0;i<max; i++) {
	em = &em8300[i];

	em8300_i2c_exit(em);

	UCREGV(Q_IrqMask)=0;
	UCREGV(Q_IrqStatus)=0;
	em->mem[0x2000]=0;
	
	em8300_free_fifo(&em->mvfifo);
	em8300_free_fifo(&em->mafifo);
	em8300_free_fifo(&em->spfifo);
	
	/* free it */
	free_irq(em->dev->irq,em);
	
	/* unmap and free memory */
	if (em->mem)
	    iounmap((unsigned *)em->mem);
    }
}

static int find_em8300(void)
{
    struct pci_dev *dev = NULL;
    unsigned char revision;
    struct em8300_s *em;
    int em8300_n=0;
    int result;
     
    if (!pcibios_present()) {
	printk(KERN_DEBUG "em8300: PCI-BIOS not present or not accessible!\n");
	return -1;
    }
    
    while ((dev = pci_find_device(PCI_VENDOR_ID_SIGMADESIGNS,PCI_DEVICE_ID_SIGMADESIGNS_EM8300, dev))) {
	em = &em8300[em8300_n];
	em->dev= dev;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	em->adr = dev->base_address[0];
	if (remap[em8300_n]) {
	    uint dw;

	    if (remap[em8300_n] < 0x1000)
		remap[em8300_n]<<=20;
	    remap[em8300_n] &= PCI_BASE_ADDRESS_MEM_MASK;
	    printk(KERN_INFO "Remapping to : 0x%08x.\n", remap[em8300_n]);
	    remap[em8300_n] |= em->adr&(~PCI_BASE_ADDRESS_MEM_MASK);
	    pci_write_config_dword(dev, PCI_BASE_ADDRESS_0, remap[em8300_n]);
	    /* commit to PCI bus */
	    pci_read_config_dword(dev,  PCI_BASE_ADDRESS_0, &dw);
	    dev->base_address[0] = em->adr;
	}
	em->adr &= PCI_BASE_ADDRESS_MEM_MASK;
#else
	em->adr = dev->resource[0].start;
#endif	
	em->memsize = 1024*1024;
	
	pci_read_config_byte(dev, PCI_CLASS_REVISION, &revision);
	printk(KERN_INFO "em8300: EM8300 %x (rev %d) ",
	       dev->device, revision);
	printk("bus: %d, devfn: %d, irq: %d, ",
	       dev->bus->number, dev->devfn, dev->irq);
	printk("memory: 0x%08lx.\n", em->adr);
	
	em->mem = ioremap(em->adr, em->memsize);
	DEBUG(printk(KERN_DEBUG "em8300: mapped-memory at 0x%p\n",em->mem));

	result = request_irq(dev->irq, em8300_irq,
			     SA_SHIRQ|SA_INTERRUPT,"em8300",(void *)em);
	
	if (result==-EINVAL) {
	    printk(KERN_ERR "em8300: Bad irq number or handler\n");
	    return -EINVAL;
	}	
	
	pci_set_master(dev);
	
	em8300_n++;
    }

    if(em8300_n)
	printk(KERN_INFO "em8300: %d EM8300 card(s) found.\n",em8300_n);
  
    return em8300_n;
}

static
int prepare_ucode_upload(struct em8300_s *em)
{
    em->mem[0x30000] = 0x1ff00;
    em->mem[0x1f50] = 0x123;

     em->mem[0x20001] = 0x0;
     em->mem[0x2000] = 0x2;
     em->mem[0x2000] = 0x0;
     em->mem[0x1ff8] = 0xffff;
     em->mem[0x1ff9] = 0xffff;
     em->mem[0x1ff8] = 0xff00;
     em->mem[0x1ff9] = 0xff00;
     em->mem[0x1c04] = 0x825a;
     em->mem[0x1c00] = 0x800;
     em->mem[0x1c04] = 0x25a;
     em->mem[0x1c08];
     em->mem[0x1c10] = 0x8;
     em->mem[0x1c20] = 0x8;
     em->mem[0x1c30] = 0x8;
     em->mem[0x1c40] = 0x8;
     em->mem[0x1c50] = 0x8;
     em->mem[0x1c60] = 0x8;
     em->mem[0x1c70] = 0x8;
     em->mem[0x1c80] = 0x8;
     em->mem[0x1c90] = 0x10;
     em->mem[0x1ca0] = 0x10;
     em->mem[0x1cb0] = 0x8;
     em->mem[0x1cc0] = 0x8;
     em->mem[0x1cd0] = 0x8;
     em->mem[0x1ce0] = 0x8;
     em->mem[0x1c01] = 0x5555;
     em->mem[0x1c02] = 0x55a;
     em->mem[0x1c03] = 0x0;
    
    return 0;
}

static
int set_regblock(struct em8300_s *em, int offset, int val, int len) {
    int i;

    em->mem[0x1c11] = offset & 0xffff;
    em->mem[0x1c12] = (offset >> 16) & 0xffff;
    em->mem[0x1c13] = len;
    em->mem[0x1c14] = len;
    em->mem[0x1c15] = 0;
    em->mem[0x1c16] = 1;
    em->mem[0x1c17] = 1;
    em->mem[0x1c18] = offset & 0xffff;
    em->mem[0x1c19] = (offset >> 16) & 0xffff;

    em->mem[0x1c1a] = 1;

    for(i=0; i < len; i+=4) 
	em->mem[0x11800] = val;
	
    if(em8300_waitfor(em,0x1c1a, 0, 1))
	return -ETIME;
    return 0;
}

static
int upload_ucode_block(struct em8300_s *em, int blocktype, int offset, int len, unsigned char *buf) {
    int i,val;

    switch(blocktype) {
    case 4:
	offset *= 2;
	em->mem[0x1c11] = offset & 0xffff;
	em->mem[0x1c12] = (offset >> 16) & 0xffff;
	em->mem[0x1c13] = len;
	em->mem[0x1c14] = len;
	em->mem[0x1c15] = 0;
	em->mem[0x1c16] = 1;
	em->mem[0x1c17] = 1;
	em->mem[0x1c18] = offset & 0xffff;
	em->mem[0x1c19] = (offset >> 16) & 0xffff;

	em->mem[0x1c1a] = 1;

	for(i=0; i < len; i+=4) {
	    val = (buf[i+2] << 24) | (buf[i+3] << 16) | (buf[i] << 8) | buf[i+1];
	    em->mem[0x11800] = val;
	}
	
	if(em8300_waitfor(em,0x1c1a, 0, 1))
	    return -ETIME;
	break;
    case 1:
	for(i=0; i < len; i+=4) {
	    val = (buf[i+1] << 24) | (buf[i] << 16) | (buf[i+3] << 8) | buf[i+2];
	    em->mem[offset/2 + i/4] = val;
	}
	break;
    case 2:
	for(i=0; i < len; i+=2) {
	    val = (buf[i+1] << 8) | buf[i];
	    em->mem[0x1000 + offset + i/2] = val;
	}
	break;
    }
    return 0;
}

static
int upload_ucode(struct em8300_s *em, void *ucode_user, int ucode_size) {
    int flags,offset,len;
    unsigned char *ucode,*p;
    int memcount,i;
    char regname[128];

    ucode = kmalloc(ucode_size, GFP_KERNEL);
    if(!ucode)
	return -ENOMEM;

    copy_from_user(ucode,ucode_user,ucode_size);

    memcount=0;

    p = ucode;
    while(memcount < ucode_size) {
	flags =  p[0] | (p[1] << 8); p+=2;
	offset = p[0] | (p[1] << 8) | (p[2] << 16) | (p[3] << 24); p+=4; 
	len =    p[0] | (p[1] << 8) | (p[2] << 16) | (p[3] << 24); p+=4;
	memcount+=10;
	len *= 2;

	if(!flags)
	    break;
	
	switch(flags & 0xf00) {
	case 0:
	    upload_ucode_block(em, flags, offset, len, p);
	    break;
	case 0x200:
	    for(i=0;i < len; i++) {
		if(p[i]) {
		    regname[i] = p[i] ^ 0xff;
		} else
		    break;
	    }
	    regname[i]=0;

	    for(i=0; i < MAX_UCODE_REGISTER; i++) {
		if(!strcmp(ucodereg_names[i], regname)) {
		    em->ucode_regs[i] = 0x1000 + offset;
		    break;
		}
	    }
	    break;
	}
	memcount += len;
	p += len;
    }
    
    kfree(ucode);
    return 0;
}

struct dicom_s dicom_PAL = 
{
    /* Luma */		  0x4000,
    /* Chroma */	  0x6060,
    /* Frame top */	  0x2e,	
    /* Frame bottom */    0x26d,
    /* Frame left   */    0x8a,
    /* Frame right  */    0x359,
    /* Visible top */	  0x2e,	
    /* Visible bottom */  0x26d,
    /* Visible left   */  0x8a,
    /* Visible right  */  0x359,
    /* TV Out         */  0x4001	
};

static
void video_setspeed(struct em8300_s *em, int speed)
{
    if(UCREGV(MicroCodeVersion) >= 0x29)
	UCREGV(MV_SCRSpeed) = speed;
    else {
	UCREGV(MV_SCRSpeed) = speed >> 8;
    }
}

static
int video_setup(struct em8300_s *em) {
    em->clockgen = 0xb | CLOCKGEN_SAMPFREQ_48;
    
    em8300_clockgen_write(em,em->clockgen);
    
    REG(0x1f47)=0x0;
    REG(0x1f5e)=0x9afe;

    UCREGV(DICOM_Control)=0x9afe;

    udelay(100);

    UCREGV(DICOM_UpdateFlag)=0x0;
    UCREGV(DICOM_VisibleLeft)=0x168;
    UCREGV(DICOM_VisibleTop)=0x2e;
    UCREGV(DICOM_VisibleRight)=0x36b;
    UCREGV(DICOM_VisibleBottom)=0x11e;
    UCREGV(DICOM_FrameLeft)=0x168;
    UCREGV(DICOM_FrameTop)=0x2e;
    UCREGV(DICOM_FrameRight)=0x36b;
    UCREGV(DICOM_FrameBottom)=0x11e;
    UCREGV(DICOM_TvOut)=0x4000;
    UCREGV(DICOM_UpdateFlag)=0x1;

    DEBUG(printk("em8300: register 0x1f4b = %x (0x138)\n",REG(0x1f4b)));
	  
    UCREGV(DICOM_Kmin)=0x455;
    REG(0x1f47)=0x18;
    REG(0x1f5e)=0x9afe;
    UCREGV(DICOM_Control)=0x9afe;

    udelay(100);
    
    UCREGV(ForcedLeftParity)=0x2;
    UCREGV(MV_Threshold)=0x50;
    REG(0x1ffa)=0x2;
    UCREGV(Q_IrqMask)=0x0;
    UCREGV(Q_IrqStatus)=0x0;
    UCREGV(Q_IntCnt)=0x64;
    REG(0x1ffb)=0xce4;
    UCREGV(MA_Threshold)=0x8;

    /* Release reset */
    REG(0x2000)=0x1;

    if(mpegvideo_command(em,0x11)) {
	DEBUG( printk("em8300: mpegvideo_command(0x11) failed\n"));
	return -ETIME;
    }

    UCREGV(SP_Status)=0x0;
    
    if(mpegvideo_command(em,0x10)) {
	DEBUG( printk("em8300: mpegvideo_command(0x10) failed\n"));
	return -ETIME;
    }

    video_setspeed(em,0x900);

    UCREGV(MV_FrameEventLo)=0xffff;
    UCREGV(MV_FrameEventHi)=0x7fff;

    em->dicom = dicom_PAL;

    if(dicom_update(em)) {
	DEBUG( printk("em8300: DICOM Update failed\n"));
	return -ETIME;
    }
    
    return 0;
}

static
int video_start(struct em8300_s *em) {
    em->irqmask |= IRQSTATUS_VIDEO_FIFO;
    UCREGV(Q_IrqMask) = em->irqmask;

    if(mpegvideo_command(em,0x11))
	return -ETIME;

    if(mpegvideo_command(em,0x10))
	return -ETIME;

    video_setspeed(em,0x900);

    UCREGV(MV_FrameEventLo)=0xffff;
    UCREGV(MV_FrameEventHi)=0x7fff;

    return mpegvideo_command(em,MVCOMMAND_START);
}

static
int video_stop(struct em8300_s *em) {
    em->irqmask &= ~IRQSTATUS_VIDEO_FIFO;
    UCREGV(Q_IrqMask) = em->irqmask;
    return mpegvideo_command(em,MVCOMMAND_STOP);
}

static
int audio_setup(struct em8300_s *em) {
    set_regblock(em, UCREGV(MA_BuffStart_Lo) |
		 (UCREGV(MA_BuffStart_Hi) << 16), 0,
		 UCREGV(MA_BuffSize));
    
    UCREGV(MA_Threshold) = 6;
    
    mpegaudio_command(em,2);
    REG(0x1f47) = 0x18;
    mpegaudio_command(em,1);
    REG(0x1fb0) = 0x62;
    set_regblock(em, 0x51c0,0,0x600);
    return 0;
}

int em8300_io_ioctl(struct inode* inode, struct file* filp, unsigned int cmd, unsigned long arg)
{
    struct em8300_s *em = filp->private_data;
    int subdevice = MINOR(inode->i_rdev) & 3;

    microcode_t uc;
    register_t reg;
    
    struct timeval tv;
    char tmpstr[1024];
    int ret,err,len;
    long tdiff,frames,scr;

    switch(subdevice) {
    case EM8300_SUBDEVICE_AUDIO:
	return em8300_audio_ioctl(em,cmd,arg);
    case EM8300_SUBDEVICE_VIDEO:
	return -EINVAL;
    case EM8300_SUBDEVICE_SUBPICTURE:
	return -EINVAL;
    }
    
    if (_IOC_DIR(cmd) != 0) {
	len = _IOC_SIZE(cmd);
	if (len < 1 || len > 65536 || arg == 0)
	    return -EFAULT;
	if (_IOC_DIR(cmd) & _IOC_WRITE)
	    if ((err = verify_area(VERIFY_READ, (void *)arg, len)) < 0)
		return err;
	if (_IOC_DIR(cmd) & _IOC_READ)
	    if ((err = verify_area(VERIFY_WRITE, (void *)arg, len)) < 0)
                                return err;
    }	
    
    switch(_IOC_NR(cmd)) {
   /* Microcode upload */
    case 0:
	copy_from_user(&uc, (void *)arg, sizeof(microcode_t));

	if((ret=prepare_ucode_upload(em)))
	    return ret;
	
	if((ret=upload_ucode(em,uc.ucode, uc.ucode_size)))
	    return ret;

	if((ret=set_regblock(em, UCREGV(MA_BuffStart_Lo) |
			     (UCREGV(MA_BuffStart_Hi) << 16),0,
			     UCREGV(MA_BuffSize))))
	    return ret;

	if((ret=video_setup(em)))
	    return ret;

	em8300_init_fifo(em,&em->mvfifo,
			 MV_PCIStart, MV_PCIWrPtr, MV_PCIRdPtr,
			 MV_PCISize, 0x900, FIFOTYPE_VIDEO);

	em8300_init_fifo(em,&em->mafifo, 
			 MA_PCIStart, MA_PCIWrPtr, MA_PCIRdPtr, 
			 MA_PCISize, 0x1000, FIFOTYPE_AUDIO); 

	em8300_init_fifo(em,&em->spfifo, 
			 SP_PCIStart, SP_PCIWrPtr, SP_PCIRdPtr, 
			 SP_PCISize, 0x900, FIFOTYPE_VIDEO); 

	if((ret=audio_setup(em)))
	    return ret;

	em->ucodeloaded = 1;

	printk(KERN_INFO "em8300: Microcode version 0x%02x loaded\n",UCREGV(MicroCodeVersion));
	
	break;
    /* Write register */
    case 1:
	copy_from_user(&reg, (void *)arg, sizeof(register_t));
	REG(reg.reg) = reg.val;
	break;
    /* Read register */
    case 2:
	copy_from_user(&reg, (void *)arg, sizeof(register_t));
	reg.val = REG(reg.reg);
	copy_to_user((void *)arg,&reg, sizeof(register_t));
	break;
    /* Get status report */
    case 3:
	frames = (UCREGV(MV_FrameCntHi) << 16) | UCREGV(MV_FrameCntLo);
	scr = (UCREGV(MV_SCRhi) << 16) | UCREGV(MV_SCRlo);
	
	do_gettimeofday(&tv);
	tdiff = TIMEDIFF(tv,em->last_status_time);
	em->last_status_time = tv;
	sprintf(tmpstr,"Time elapsed: %ld us\nIRQ time period: %ld\nInterrupts : %d\nFrames: %ld\nSCR: %ld\n",
		tdiff, em->irqtimediff,em->irqcount,frames-em->frames,scr-em->scr);
	em->irqcount = 0;
	em->frames = frames;
	em->scr = scr;
	copy_to_user((void *)arg,tmpstr,strlen(tmpstr)+1);
	break;
    default:
	return -EINVAL;
    }
    return 0;
}

static
int em8300_io_open(struct inode* inode, struct file* filp) 
{
    int card = MINOR(inode->i_rdev)/4;
    int subdevice = MINOR(inode->i_rdev) & 3;
    struct em8300_s *em = &em8300[card];
    int err=0;
  
    if(card >= em8300_cards)
	return -ENODEV;

    if(em8300[card].inuse[subdevice])
	return -EBUSY;
  
    filp->private_data = &em8300[card];

    switch(subdevice) {
    case EM8300_SUBDEVICE_CONTROL:
	break;
    case EM8300_SUBDEVICE_AUDIO:
	err =  em8300_audio_open(em);
	break;
    case EM8300_SUBDEVICE_VIDEO:
	if(!em->ucodeloaded)
	    return -ENODEV;
    	err = video_start(em);
	break;
    default:
	return -ENODEV;
	break;
    }

    if(err)
	return err;
    
    em8300[card].inuse[subdevice] = 1;

    MOD_INC_USE_COUNT;
  
    return(0);
}

int em8300_io_write(struct file *file, const char * buf,
		    size_t count, loff_t *ppos)
{
    struct em8300_s *em = file->private_data;
    int total_bytes_written=0,copy_size;
    struct fifo_s *fifo;
    int subdevice = MINOR(file->f_dentry->d_inode->i_rdev) & 3;
    int swapbyteorder=0;
    
    switch (subdevice) {
    case EM8300_SUBDEVICE_VIDEO:
	fifo = &em->mvfifo;
	break;
    case EM8300_SUBDEVICE_AUDIO:
	fifo = &em->mafifo;
	swapbyteorder = em->swapbytes;
	break;
    case EM8300_SUBDEVICE_SUBPICTURE:
	fifo = &em->spfifo;
	break;
    default:
	return -EPERM;
    }

    if(!fifo->valid)
	return -EPERM;	
    
    while(count) {
	copy_size = em8300_write_fifo(fifo, count, buf, swapbyteorder);

	if(copy_size < 0)
	    return -EIO;
	
	count -= copy_size;
	buf += copy_size;
	total_bytes_written += copy_size;
	
	if(count) {
	    fifo->waiting=1;
	    interruptible_sleep_on(&fifo->wait);
	    fifo->waiting=0;
	}
	
	if(signal_pending(current)) {
	    if(total_bytes_written) 
		return total_bytes_written;
	    else
		return -EINTR;
	}
    }
    return total_bytes_written;
    
}

int em8300_io_mmap(struct file *file, struct vm_area_struct *vma)
{
    struct em8300_s *em = file->private_data;
    unsigned long size = vma->vm_end - vma->vm_start;
    int subdevice = MINOR(file->f_dentry->d_inode->i_rdev) & 3;

    if(subdevice != EM8300_SUBDEVICE_CONTROL)
	return -EPERM;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
    switch(vma->vm_offset) {
#else	
    switch(vma->vm_pgoff) {
#endif	
    case 0:
	if(size > em->memsize)
	    return -EINVAL;
	
	remap_page_range(vma->vm_start,
			 em->adr,
			 vma->vm_end - vma->vm_start,
			 vma->vm_page_prot);
	vma->vm_file = file;
	file->f_dentry->d_inode->i_count++;
	break;
    default:
	return -EINVAL;
    }
    return 0;
}

int em8300_io_release(struct inode* inode, struct file* filp) 
{
    struct em8300_s *em = filp->private_data;
    int subdevice = MINOR(inode->i_rdev) & 3;
    
    switch(subdevice) {
    case EM8300_SUBDEVICE_AUDIO:
	em8300_audio_release(em);
	break;
    case EM8300_SUBDEVICE_VIDEO:
	video_stop(em);
	break;
    }
    
    em->inuse[subdevice]=0;
    
    MOD_DEC_USE_COUNT;

    return(0);
}
   

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
static struct file_operations em8300_fops = {
  NULL,
  NULL,
  em8300_io_write,
  NULL,
  NULL,
  em8300_io_ioctl,
  em8300_io_mmap,
  em8300_io_open,
  NULL,
  em8300_io_release,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL };
#else
static struct file_operations em8300_fops = {
  NULL,
  NULL,
  em8300_io_write,
  NULL,
  NULL,
  em8300_io_ioctl,
  em8300_io_mmap,
  em8300_io_open,
  NULL,
  em8300_io_release,
  NULL,
  NULL,
  NULL };
#endif

void cleanup_module(void) {
    release_em8300(em8300_cards);
    unregister_chrdev(EM8300_MAJOR,EM8300_LOGNAME);
}

int init_module(void)
{
    int card;
    struct em8300_s *em;

    memset(&em8300_cards, 0, sizeof(em8300_cards));
    
    /* Find EM8300 cards */
    em8300_cards = find_em8300();

    /* Initialize EM8300 cards */
    for(card = 0; card < em8300_cards; card++) {
	em = &em8300[card];

	em->irqmask = 0;
	
	em->encoder = NULL;
	em->eeprom = NULL;
	
	em8300_i2c_init(em);

	em->linecounter=0;
    }
    if(register_chrdev(EM8300_MAJOR, EM8300_LOGNAME, &em8300_fops)) {
	printk(KERN_ERR "em8300: unable to get major %d\n", EM8300_MAJOR);
	return -EBUSY;
    }
    
    return 0;
}

